# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt

# 표준편차 sigma, 평균 mu인 정규분포에서 3개를 추출
# 각각 w_1, w_2, b임
def Create_genome(sigma=1, mu=1):
    value = sigma*np.random.randn(3) + mu
    return value

# 유전 객체 4개 만들기
def Create_4genome(sigma=1, mu=1):
    gen_list = np.array([])
    for i in range(4):
        value = Create_genome(sigma, mu)
        gen_list = np.concatenate([gen_list, value], axis=0)
    gen_list = np.reshape(gen_list, (-1, 3))
    return gen_list

# 적합도 계산 : |target - x_1w_1 + x_2w_2 + b|
def Appropriate(list_gen, x, y):
  appro_list = np.array([])
  for gen in list_gen:
    w = gen[:2]
    b = gen[2]
    appro = np.array([np.abs(y - np.sum(x@w + b))])
    appro = np.round(appro, 2)
    appro_list = np.concatenate([appro_list, appro], axis=0)
  return appro_list

# 개체 중 적합도가 가장 작은 2개의 객체 선택하기
def Select_appropriate(list_gen, x, y):
  list_appro = Appropriate(list_gen, x, y)
  argsort = list_appro.argsort()
  two_genlist = np.array([])
  two_genlist = np.concatenate([two_genlist, list_gen[argsort[0]]], axis=0)
  two_genlist = np.concatenate([two_genlist, list_gen[argsort[1]]], axis=0)
  two_genlist = np.reshape(two_genlist, (-1, 3))
  return two_genlist



#유전자 객체 교차 구현하기
def Intersect_genome(list_gen, state):
  temp_list = np.array([])
  for gen in list_gen:
    value = gen[state]
    value = np.array([value])
    temp_list = np.concatenate([temp_list, value], axis=0)
  copy_list = list_gen.copy()
  copy_list[0][state] = temp_list[1]
  copy_list[1][state] = temp_list[0]
  return copy_list


# 확률적 돌연변이, 확률에 따라 돌연변이가 발생하도록 한다.
# 돌연변이 조건에 해당하면 유전물질을 아예 새로 생성
def Mutation(list_gen, prob=0.1):
  event = np.random.choice((0, 1), p=[1-prob, prob])
  if event == True:
    mutant_list = np.array([])
    for i in range(2):
      value = Create_genome()
      mutant_list = np.concatenate([mutant_list, value], axis=0)
    
    mutant_list = np.reshape(mutant_list, (-1, 3))
    return mutant_list
  else:
    return list_gen

# 4개의 객체 중 적합도가 가장 작은 유전 객체 2개 
# + 유전 객체 2개 교차 구현한 것 합치기
def Combine_genome(list_1, list_2):
  return np.concatenate((list_1, list_2), axis=0)

def Fit(genlist_four, x, y, epochs, prob=0.1, period=None):
    for epoch in range(epochs):
        genlist_two= Select_appropriate(genlist_four, x, y)
        change_num = np.random.choice(3, 1)
        intersect_gen = Intersect_genome(genlist_two, change_num[0]) 
        intersect_gen = Mutation(intersect_gen, prob)
        genlist_four = Combine_genome(genlist_two, intersect_gen)  
        # period 매개변수에 따라 epoch 출력
        if period is not None and (epoch + 1) % period == 0:
            appr = Appropriate(genlist_four, x, y)
            print(f"Generation: {epoch + 1}, Appropriate: {appr}")
            Display_genome(genlist_four, target=y, appr=appr, epoch=epoch+1)
    print('Complete!')
    return genlist_four

def Display_genome(genome, target, appr=None, epoch = None):
    if genome.ndim != 2:
        genome = np.expand_dims(genome, axis=0)
    
    # genome 갯수 계산
    num_genome = genome.shape[0]
    
    # 유전자 값을 세로 막대 그래프로 표현
    plt.figure(figsize=(2*num_genome, 4))
        
    # 파스텔 톤의 RGB 색상 사용
    green = (0.8, 0.9, 0.8)  # 연한 녹색
    violet = (0.9, 0.8, 0.9)  # 연한 보라색
    blue = (0.8, 0.8, 0.9)  # 연한 파란색
        
    # 4개의 유전자를 subplot으로 표현
    for i, genes in enumerate(genome):
        plt.subplot(1, num_genome, i+1)
            
        gene1, gene2, gene3 = genes
            
        # 세로 축 최대 높이를 20으로 설정
        plt.ylim(0, 20)
            
        plt.bar(0, gene1, color=green, width=0.5)
        plt.bar(0, gene2, bottom=gene1, color=violet, width=0.5)
        plt.bar(0, gene3, bottom=gene1+gene2, color=blue, width=0.5)
            
        # 각 유전자 값을 텍스트로 표시
        plt.text(0.15, gene1/2, f"w1{gene1:.2f}", va='center', ha='right', color='k')
        plt.text(0.15, gene1+gene2/2, f"w2: {gene2:.2f}", va='center', ha='right', color='k')
        plt.text(0.06, gene1+gene2+gene3/2, f"b: {gene3:.2f}", va='center', ha='right', color='k')
            
    if appr is not None:
        for j in range(num_genome):
            plt.subplot(1, num_genome, j+1)
            plt.text(-0.23, -1, f"Appropriate: {appr[j]}", va='top', ha='left', color='k')    
    if epoch is not None:
        plt.suptitle(f"Epochs: {epoch}", fontsize=16, position = (0.5, 1))
            
    plt.tight_layout()
    plt.show()


# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt

# (4, 4)의 유전 객체 생성
def Create_genome():
  return np.random.choice((0, 1), (4, 4))

# (4, 4)의 유전 객체 4개 생성
def Create_4genome():
    gen_list = Create_genome()
    for i in range(3):
        value = Create_genome()
        gen_list = np.vstack((gen_list, value))
    gen_list = np.reshape(gen_list, (-1, 4, 4))
    return gen_list

# 목표값(target)에 대한 적합도 계산
def Appropriate(list_gen, target):
  appro_list = np.array([])
  for genome in list_gen:
    appro = np.array([np.sum(np.abs(target-genome))])
    appro_list = np.concatenate([appro_list, appro], axis=0)
  return appro_list

# 개체 중 적합도가 가장 작은 2개의 객체 선택하기
def Select_appropriate(list_gen, target):
  list_appro = Appropriate(list_gen, target) # 적합도 배열 
  argsort = list_appro.argsort() # 적합도 배열 인덱스 오름차순 정렬
  select_one = list_gen[argsort[0]] # 적합도가 가장 작은 객체
  select_two = list_gen[argsort[1]] # 적합도가 두 번째로 작은 객체
  two_genlist = np.vstack((select_one, select_two)) # 적합도 객체 쌓기 (8, 4)
  two_genlist = np.reshape(two_genlist, (-1, 4, 4)) # shape 재정의 (2, 4, 4)
  return two_genlist

# 감수 분열을 이용한 교차 구현
# 감수 분열을 하면 행 1-2(인덱스 0, 1) , 행 3-4(인덱스 2, 3)로 분열된다. 
# 서로 교차할 물질은 행 3-4번 끼리 서로 교체함.
def Intersect_genome(list_gen):
  temp_list = np.vstack((list_gen[0][2:4], list_gen[1][2:4]))
  copy_list = list_gen.copy()
  copy_list[0][2:4] = temp_list[2:4]
  copy_list[1][2:4] = temp_list[0:2]
  return copy_list

# 돌연변이가 발생하면 위의 행 2개 또는 아래 행 2개를 무작위의 0, 1로 변경됨.
def Mutation(list_gen, prob=0.1):
  event = np.random.choice((0, 1), p=[1-prob, prob])
  if event == True:
    mutant_list = np.random.choice((0, 1), (4, 4))
    copy_list = list_gen.copy()
    high_low = np.random.choice((0, 1), p=[0.5, 0.5])
    if high_low == True:
      copy_list[0][0:2] = mutant_list[0:2]
      copy_list[1][0:2] = mutant_list[2:4]
    else:
      copy_list[0][2:4] = mutant_list[0:2]
      copy_list[1][2:4] = mutant_list[2:4]
    return copy_list
  else:
    return list_gen

# 4개의 객체 중 적합도가 가장 작은 유전 객체 2개 + 유전 객체 2개 교차 구현한 것 합치기
def Combine_genome(list_1, list_2):
  return np.concatenate((list_1, list_2))

def Fit(genlist_four, target, epochs, prob=0.1, period=None):
    for epoch in range(epochs):
        genlist_two= Select_appropriate(genlist_four, target)
        intersect_gen = Intersect_genome(genlist_two) 
        intersect_gen = Mutation(intersect_gen, prob=0.1)
        genlist_four = Combine_genome(genlist_two, intersect_gen) 
        
        # period 매개변수에 따라 epoch 출력
        if period is not None and (epoch + 1) % period == 0:
            appr = Appropriate(genlist_four, target)
            print(f"Generation: {epoch + 1}, Appropriate: {appr}")
            Display_image(genlist_four, appr, epoch+1)
    print('Complete!')    
    return genlist_four
    
# 이미지 출력 기능
def Display_image(image, appr=None, epoch=None):
    
    # 이미지가 3차원이 아니라면 (1개라면), 3차원으로 수정
    if image.ndim != 3:
        image = np.expand_dims(image, axis=0)
    
    # 이미지 개수 계산
    num_image = image.shape[0]
    
    plt.figure(figsize=(3*num_image, 6))
    for i in range(num_image):
        plt.subplot(1, num_image, i+1)
        plt.xticks(np.arange(0, 4, 1))
        plt.yticks(np.arange(0, 4, 1))
        plt.imshow(image[i], cmap='gray')
    if epoch is not None:
        plt.suptitle(f"Generation: {epoch}", fontsize=16, position = (0.5, 0.75))
    if appr is not None:
        for j in range(num_image):
            plt.subplot(1, num_image, j+1)
            plt.text(0.5, -0.2, f"Appropriate: {appr[j]}", ha='center', transform=plt.gca().transAxes)
    plt.show()



# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt

dice = np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=np.int32)

#주사위 던지기 -> np.array 1_dimension
def Throw_dice():
  value = np.random.choice(dice, 1)
  return value

#유전자 만들기 -> np.array 1_dimension 3 element
def Create_genome():
  genome = np.array([])
  for i in range(3):
    value = Throw_dice()
    genome = np.concatenate([genome, value], axis=0)
  return genome

#유전자 만들기 -> np.array 2_dimension (4, 3 element)
def Create_4genome():
    gen_list = np.array([])
    for i in range(4):
        value = Create_genome()
        gen_list = np.concatenate([gen_list, value], axis=0)
    gen_list = np.reshape(gen_list, (-1, 3))
    return gen_list

# 적합도 계산 -> 적합도는 |20-유전자|로 계산됨. np.array 1_dimension
def Appropriate(list_gen):
  appro_list = np.array([])
  for gen in list_gen:
    gen_sum = np.sum(gen)
    appro = np.array([np.abs(20-gen_sum)])
    appro_list = np.concatenate([appro_list, appro], axis=0)
  return appro_list

# 개체 중 적합도가 가장 작은 2개의 객체 선택하기
def Select_appropriate(list_gen):
  list_appro = Appropriate(list_gen)
  argsort = list_appro.argsort()
  two_genlist = np.array([])
  two_genlist = np.concatenate([two_genlist, list_gen[argsort[0]]], axis=0)
  two_genlist = np.concatenate([two_genlist, list_gen[argsort[1]]], axis=0)
  two_genlist = np.reshape(two_genlist, (-1, 3))
  return two_genlist

#유전자 객체 교차 구현하기
def Intersect_genome(list_gen, state):
  temp_list = np.array([])
  for gen in list_gen:
    value = gen[state]
    value = np.array([value])
    temp_list = np.concatenate([temp_list, value], axis=0)
  
  copy_list = list_gen.copy()
  copy_list[0][state] = temp_list[1]
  copy_list[1][state] = temp_list[0]
  return copy_list

# 4개의 객체 중 적합도가 가장 작은 유전 객체 2개 + 유전 객체 2개 교차 구현한 것 합치기
def Combine_genome(list_1, list_2):
  return np.concatenate((list_1, list_2), axis=0)

# 주사위 숫자가 5일 때 돌연변이 일으키기
# 돌연변이를 일으킬 위치(인덱스 0, 1, 2)를 랜덤하게 생성
def Mutation(list_gen, prob=0.1):
  mutant_list = list_gen.copy()
  event = np.random.choice((0, 1), p=[1-prob, prob])
  if event == True :
    mutant_state = np.random.choice(np.arange(0, 3), 1)
    for mutant in mutant_list:
      mutant_dice = Throw_dice()
      mutant[mutant_state[0]] = mutant_dice[0]
  return mutant_list

def Fit(genlist_four, epochs, prob=0.1, period=None):
    for epoch in range(epochs):
        genlist_two= Select_appropriate(genlist_four)
        change_num = np.random.choice(3, 1)
        intersect_gen = Intersect_genome(genlist_two, change_num[0]) 
        intersect_gen = Mutation(intersect_gen, prob)
        genlist_four = Combine_genome(genlist_two, intersect_gen) 
        # period 매개변수에 따라 epoch 출력
        if period is not None and (epoch + 1) % period == 0:
            appr = Appropriate(genlist_four)
            print(f"Generation: {epoch + 1}, Appropriate: {appr}")
            Display_genome(genlist_four, appr=appr, epoch=epoch+1)
    print('Complete!')    
    return genlist_four
    
def Display_genome(genome, appr=None, epoch = None):
    
    if genome.ndim != 2:
        genome = np.expand_dims(genome, axis=0)
    # genome 갯수 계산
    num_genome = genome.shape[0]
    
    # 유전자 값을 세로 막대 그래프로 표현
    plt.figure(figsize=(2*num_genome, 4))
        
    # 파스텔 톤의 RGB 색상 사용
    red = (1, 0.7, 0.7)
    green = (0.7, 1, 0.7)
    blue = (0.7, 0.7, 1)
    
    # 1~4개의 유전자를 subplot으로 표현
    for i, genes in enumerate(genome):
        plt.subplot(1, num_genome, i+1)
            
        gene1, gene2, gene3 = genes
            
        # 세로 축 최대 높이를 20으로 설정
        plt.ylim(0, 20)
            
        plt.bar(0, gene1, color=red, width=0.5)
        plt.bar(0, gene2, bottom=gene1, color=green, width=0.5)
        plt.bar(0, gene3, bottom=gene1+gene2, color=blue, width=0.5)
        
        plt.text(0.07, gene1/2, f"{gene1:.2f}", va='center', ha='right', color='k')
        plt.text(0.07, gene1+gene2/2, f"{gene2:.2f}", va='center', ha='right', color='k')
        plt.text(0.07, gene1+gene2+gene3/2, f"{gene3:.2f}", va='center', ha='right', color='k')    
       
        plt.xticks([])
        plt.yticks([0, 5, 10, 15, 20])
    if appr is not None:
        for j in range(num_genome):
            plt.subplot(1, num_genome, j+1)
            plt.text(-0.23, -1, f"Appropriate: {appr[j]}", va='top', ha='left', color='k')

    if epoch is not None:
        plt.suptitle(f"Generation: {epoch}", fontsize=16, position = (0.5, 1))
            
    plt.tight_layout()
    plt.show()


from setuptools import setup, find_packages

setup(
      name='edugenome',
      version='0.2.6',
      description='It consists of three genetic algorithms that are simply \
          implemented with Python code for genetic algorithm training: \
              creating a number sum of 20, creating (4, 4) images, and \
                  implementing linear regression.',
      author='majestyblue',
      author_email = 'majestyblue88@gmail.com',
      url = 'https://github.com/majestyblue/edugenome',
      install_requires=['numpy', 'matplotlib'],
      packages=find_packages(exclude=[]),
      keywords=['education', 'genome', 'genetic algorithm'],
      python_requires='>=3.6',
      package_data={},
      zip_safe=False
)